/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * include/SIL/sysdep/psp/common.h: Global declarations specific to the PSP.
 */

#ifndef SIL_SYSDEP_PSP_COMMON_H
#define SIL_SYSDEP_PSP_COMMON_H

EXTERN_C_BEGIN

/*************************************************************************/
/*************************************************************************/

/* Icon size used for save files.  (The image data passed to
 * userdata_save_savefile() must be of exactly this size.) */

#define PSP_SAVE_IMAGE_WIDTH  144
#define PSP_SAVE_IMAGE_HEIGHT  80

/*-----------------------------------------------------------------------*/

/* Define malloc() and operator new/delete replacements which use our
 * custom allocation functions (with debugging support, if appropriate). */

#ifndef SIL_MEMORY_FORBID_MALLOC

extern void *malloc(size_t size);
extern void *calloc(size_t nmemb, size_t size);
extern void *realloc(void *ptr, size_t size);
extern void free(void *ptr);

#ifdef DEBUG

extern void *debug_malloc(size_t size, const char *file, int line);
extern void *debug_calloc(size_t nmemb, size_t size, const char *file, int line);
extern void *debug_realloc(void *ptr, size_t size, const char *file, int line);
extern void debug_free(void *ptr, const char *file, int line);
/* Redefine malloc(), etc. to point to the debug functions. */
# define malloc(size)        debug_malloc(size, __FILE__, __LINE__)
# define calloc(nmemb,size)  debug_calloc(nmemb, size, __FILE__, __LINE__)
# define realloc(ptr,size)   debug_realloc(ptr, size, __FILE__, __LINE__)
# define free(ptr)           debug_free(ptr, __FILE__, __LINE__)

#ifdef __cplusplus
EXTERN_C_END

/* First include some system headers that use weird new() syntax, at least
 * in GCC. */
#include <algorithm>
#include <string>
#include <vector>

/* Define the debugging versions of the operators. */
inline void *operator new(size_t size, const char *file, int line) {
    return debug_malloc(size, file, line);
}
inline void *operator new[](size_t size, const char *file, int line) {
    return debug_malloc(size, file, line);
}
inline void operator delete(void *ptr, const char *file, int line) {
    return debug_free(ptr, file, line);
}
inline void operator delete[](void *ptr, const char *file, int line) {
    return debug_free(ptr, file, line);
}

/* Define a macro that converts ordinary "new" into a call to the debugging
 * version.  For some reason, this does _not_ work with "delete". */
#define new new(__FILE__,__LINE__)
//#define delete delete(__FILE__,__LINE__)

EXTERN_C_BEGIN
#endif  // __cplusplus

#endif  // DEBUG

#endif  // !SIL_MEMORY_FORBID_MALLOC

/*-----------------------------------------------------------------------*/

/* Force atof() to single precision. */
#undef atof
#define atof(s)  strtof((s), NULL)

/*-----------------------------------------------------------------------*/

/* Miscellaneous exported routines. */

/**
 * psp_movie_set_direct_audio:  Set whether movie audio should be rendered
 * through the software mixer or directly to a hardware sound channel.
 * Direct hardware rendering requires slightly less CPU processing, but
 * the movie audio may lose sync with any other audio being played at the
 * same time.  The default is to use the software mixer.
 *
 * This flag must not be changed while a movie is being played.
 *
 * [Parameters]
 *     flag: True to enable direct hardware output; false to use the
 *         software mixer.
 */
extern void psp_movie_set_direct_audio(int flag);

/**
 * psp_userdata_set_low_priority_mode:  Enable or disable the use of low
 * thread priorities for the PSP save data utility.  Enabling low-priority
 * mode will give smoother performance, but client code must take care not
 * to monopolize the CPU while the save data utility is running.
 *
 * [Parameters]
 *     on: True to enable low-priority mode, false to disable.
 */
extern void psp_userdata_set_low_priority_mode(int on);

/**
 * psp_userdata_set_stats_file_info:  Set the title and descriptive text
 * to be stored with the PSP save data file containing statistics
 * information.
 *
 * [Parameters]
 *     title: Title text to associate with the file.
 *     desc: Descriptive text to associate with the file.
 */
extern void psp_userdata_set_stats_file_info(const char *title, const char *desc);


#ifdef DEBUG

/**
 * psp_debug_display_memory_map:  Display a map of memory usage on the top
 * of the screen, along with current usage statistics.  This display is
 * intended to be used in place of the memory meter of the standard debug
 * interface.
 *
 * This function is only available when building with DEBUG defined.
 */
extern void psp_debug_display_memory_map(void);

/**
 * psp_debug_display_ge_info:  Display rendering and buffer usage
 * statistics for the graphics subsystem.
 *
 * This function is only available when building with DEBUG defined.
 */
extern void psp_debug_display_ge_info(void);

/**
 * psp_debug_display_log:  Display the most recent debug messages as an
 * overlay on the screen.
 *
 * This function is only available when building with DEBUG defined.
 */
extern void psp_debug_display_log(void);

/**
 * psp_debug_dump_log:  Write the most recent debug messages to the given
 * file.
 *
 * This function is only available when building with DEBUG defined.
 */
extern void psp_debug_dump_log(const char *path);

#endif  // DEBUG

/*************************************************************************/
/*************************************************************************/

EXTERN_C_END

#endif  // SIL_SYSDEP_PSP_COMMON_H

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
