/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/sound/decode-ogg.c: Tests for decoding of Ogg Vorbis audio data.
 */

#ifdef INCLUDE_TESTS  // To the end of the file.

#include "src/base.h"
#include "src/memory.h"
#include "src/sound.h"
#include "src/sound/decode.h"
#include "src/sound/decode-ogg.h"
#include "src/sysdep.h"
#include "src/test/base.h"

#ifndef SIL_SOUND_INCLUDE_OGG
int test_sound_decode_ogg(void) {
    DLOG("Ogg Vorbis support disabled, nothing to test.");
    return 1;
}
#else  // To the end of the file.

#ifdef SIL_SOUND_USE_TREMOR
# include <ivorbisfile.h>
#else
# define OV_EXCLUDE_STATIC_CALLBACKS
# include <vorbis/vorbisfile.h>
#endif

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Expected PCM output from the Ogg Vorbis test files.  We need multiple
 * versions of the data because the floating-point libvorbis library and
 * integer Tremor library round differently. */

static const int16_t mono_pcm[] = {
#ifdef SIL_SOUND_USE_TREMOR
    9763, 9445, -9591, -9750, 9573, 9868,
#else
    9764, 9446, -9591, -9749, 9574, 9868,
#endif
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   // These aren't checked.
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    -10151, -9443
};

static const int16_t stereo_pcm[] = {
#ifdef SIL_SOUND_USE_TREMOR
# ifdef SIL_ARCH_ARM
    9401, -9314, 9059, -9044, 9127, -9276,
# else
    9401, -9313, 9059, -9044, 9127, -9276,
# endif
#else
    9402, -9313, 9060, -9043, 9128, -9276,
#endif
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  // These aren't checked.
#ifdef SIL_SOUND_USE_TREMOR
    -10662
#else
    -10661
#endif
};

/*************************************************************************/
/***************************** Test routines *****************************/
/*************************************************************************/

DEFINE_GENERIC_TEST_RUNNER(test_sound_decode_ogg)

/*-----------------------------------------------------------------------*/

TEST(test_decode)
{
    SysFile *file;
    ASSERT(file = sys_file_open("testdata/sound/square.ogg"));
    SoundDecodeHandle *decoder;
    CHECK_TRUE(decoder = sound_decode_open_from_file(
                   SOUND_FORMAT_OGG, file, 0, sys_file_size(file),
                   0, 0, 0, 1));
    sys_file_close(file);
    CHECK_FALSE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4000);

    int16_t pcm[38];
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 3));
    CHECK_INTEQUAL(pcm[0], mono_pcm[0]);
    CHECK_INTEQUAL(pcm[1], mono_pcm[1]);
    CHECK_INTEQUAL(pcm[2], mono_pcm[2]);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 3));
    CHECK_INTEQUAL(pcm[0], mono_pcm[3]);
    CHECK_INTEQUAL(pcm[1], mono_pcm[4]);
    CHECK_INTEQUAL(pcm[2], mono_pcm[5]);
    sound_decode_reset(decoder);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 3));
    CHECK_INTEQUAL(pcm[0], mono_pcm[0]);
    CHECK_INTEQUAL(pcm[1], mono_pcm[1]);
    CHECK_INTEQUAL(pcm[2], mono_pcm[2]);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 38));
    CHECK_INTEQUAL(pcm[36], mono_pcm[39]);
    CHECK_INTEQUAL(pcm[37], 0);

    sound_decode_close(decoder);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_decode_memory_failure)
{
    SysFile *file;
    ASSERT(file = sys_file_open("testdata/sound/square.ogg"));
    const int64_t datalen = sys_file_size(file);
    void *data;
    ASSERT(data = mem_alloc(datalen, 0, 0));
    ASSERT(sys_file_read(file, data, datalen) == datalen);
    sys_file_close(file);

    SoundDecodeHandle *decoder;
    CHECK_MEMORY_FAILURES(
        decoder = sound_decode_open(SOUND_FORMAT_OGG, data, datalen,
                                    0, 0, 0, 1));
    int16_t pcm[3];
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 3));
    CHECK_INTEQUAL(pcm[0], mono_pcm[0]);
    CHECK_INTEQUAL(pcm[1], mono_pcm[1]);
    CHECK_INTEQUAL(pcm[2], mono_pcm[2]);
    sound_decode_close(decoder);

    mem_free(data);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_decode_stereo)
{
    SysFile *file;
    ASSERT(file = sys_file_open("testdata/sound/square-stereo.ogg"));
    SoundDecodeHandle *decoder;
    CHECK_TRUE(decoder = sound_decode_open_from_file(
                   SOUND_FORMAT_OGG, file, 0, sys_file_size(file),
                   0, 0, 0, 1));
    sys_file_close(file);
    CHECK_TRUE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4000);

    int16_t pcm[36];
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 3));
    CHECK_INTEQUAL(pcm[0], stereo_pcm[0]);
    CHECK_INTEQUAL(pcm[1], stereo_pcm[0]);
    CHECK_INTEQUAL(pcm[2], stereo_pcm[1]);
    CHECK_INTEQUAL(pcm[3], stereo_pcm[1]);
    CHECK_INTEQUAL(pcm[4], stereo_pcm[2]);
    CHECK_INTEQUAL(pcm[5], stereo_pcm[2]);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 3));
    CHECK_INTEQUAL(pcm[0], stereo_pcm[3]);
    CHECK_INTEQUAL(pcm[1], stereo_pcm[3]);
    CHECK_INTEQUAL(pcm[2], stereo_pcm[4]);
    CHECK_INTEQUAL(pcm[3], stereo_pcm[4]);
    CHECK_INTEQUAL(pcm[4], stereo_pcm[5]);
    CHECK_INTEQUAL(pcm[5], stereo_pcm[5]);
    sound_decode_reset(decoder);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 3));
    CHECK_INTEQUAL(pcm[0], stereo_pcm[0]);
    CHECK_INTEQUAL(pcm[1], stereo_pcm[0]);
    CHECK_INTEQUAL(pcm[2], stereo_pcm[1]);
    CHECK_INTEQUAL(pcm[3], stereo_pcm[1]);
    CHECK_INTEQUAL(pcm[4], stereo_pcm[2]);
    CHECK_INTEQUAL(pcm[5], stereo_pcm[2]);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 18));
    CHECK_INTEQUAL(pcm[32], stereo_pcm[19]);
    CHECK_INTEQUAL(pcm[33], stereo_pcm[19]);
    CHECK_INTEQUAL(pcm[34], 0);
    CHECK_INTEQUAL(pcm[35], 0);

    sound_decode_close(decoder);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_loop)
{
    SysFile *file;
    ASSERT(file = sys_file_open("testdata/sound/square.ogg"));
    SoundDecodeHandle *decoder;
    CHECK_TRUE(decoder = sound_decode_open_from_file(
                   SOUND_FORMAT_OGG, file, 0, sys_file_size(file),
                   1, 2, 3, 1));
    sys_file_close(file);
    CHECK_FALSE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4000);

    int16_t pcm[9];
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 9));
    CHECK_INTEQUAL(pcm[0], mono_pcm[0]);
    CHECK_INTEQUAL(pcm[1], mono_pcm[1]);
    CHECK_INTEQUAL(pcm[2], mono_pcm[2]);
    CHECK_INTEQUAL(pcm[3], mono_pcm[3]);
    CHECK_INTEQUAL(pcm[4], mono_pcm[4]);
    CHECK_INTEQUAL(pcm[5], mono_pcm[2]);
    CHECK_INTEQUAL(pcm[6], mono_pcm[3]);
    CHECK_INTEQUAL(pcm[7], mono_pcm[4]);
    CHECK_INTEQUAL(pcm[8], mono_pcm[2]);

    sound_decode_close(decoder);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_loop_to_end)
{
    SysFile *file;
    ASSERT(file = sys_file_open("testdata/sound/square.ogg"));
    SoundDecodeHandle *decoder;
    CHECK_TRUE(decoder = sound_decode_open_from_file(
                   SOUND_FORMAT_OGG, file, 0, sys_file_size(file),
                   1, 2, -1, 1));
    sys_file_close(file);
    CHECK_FALSE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4000);

    int16_t pcm[43];
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 43));
    CHECK_INTEQUAL(pcm[0], mono_pcm[0]);
    CHECK_INTEQUAL(pcm[1], mono_pcm[1]);
    CHECK_INTEQUAL(pcm[2], mono_pcm[2]);
    CHECK_INTEQUAL(pcm[3], mono_pcm[3]);
    CHECK_INTEQUAL(pcm[4], mono_pcm[4]);
    CHECK_INTEQUAL(pcm[5], mono_pcm[5]);
    CHECK_INTEQUAL(pcm[38], mono_pcm[38]);
    CHECK_INTEQUAL(pcm[39], mono_pcm[39]);
    CHECK_INTEQUAL(pcm[40], mono_pcm[2]);
    CHECK_INTEQUAL(pcm[41], mono_pcm[3]);
    CHECK_INTEQUAL(pcm[42], mono_pcm[4]);

    sound_decode_close(decoder);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_loop_past_end)
{
    SysFile *file;
    ASSERT(file = sys_file_open("testdata/sound/square.ogg"));
    SoundDecodeHandle *decoder;
    CHECK_TRUE(decoder = sound_decode_open_from_file(
                   SOUND_FORMAT_OGG, file, 0, sys_file_size(file),
                   1, 2, 42, 1));
    sys_file_close(file);
    CHECK_FALSE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4000);

    int16_t pcm[43];
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 43));
    CHECK_INTEQUAL(pcm[0], mono_pcm[0]);
    CHECK_INTEQUAL(pcm[1], mono_pcm[1]);
    CHECK_INTEQUAL(pcm[2], mono_pcm[2]);
    CHECK_INTEQUAL(pcm[3], mono_pcm[3]);
    CHECK_INTEQUAL(pcm[4], mono_pcm[4]);
    CHECK_INTEQUAL(pcm[5], mono_pcm[5]);
    CHECK_INTEQUAL(pcm[38], mono_pcm[38]);
    CHECK_INTEQUAL(pcm[39], mono_pcm[39]);
    CHECK_INTEQUAL(pcm[40], mono_pcm[2]);
    CHECK_INTEQUAL(pcm[41], mono_pcm[3]);
    CHECK_INTEQUAL(pcm[42], mono_pcm[4]);

    sound_decode_close(decoder);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_loop_starts_at_end)
{
    SysFile *file;
    ASSERT(file = sys_file_open("testdata/sound/square.ogg"));
    SoundDecodeHandle *decoder;
    CHECK_TRUE(decoder = sound_decode_open_from_file(
                   SOUND_FORMAT_OGG, file, 0, sys_file_size(file),
                   1, 40, 10, 1));
    sys_file_close(file);
    CHECK_FALSE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4000);

    int16_t pcm[40];
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 40));
    CHECK_INTEQUAL(pcm[0], mono_pcm[0]);
    CHECK_INTEQUAL(pcm[1], mono_pcm[1]);
    CHECK_INTEQUAL(pcm[2], mono_pcm[2]);
    CHECK_INTEQUAL(pcm[3], mono_pcm[3]);
    CHECK_INTEQUAL(pcm[4], mono_pcm[4]);
    CHECK_INTEQUAL(pcm[5], mono_pcm[5]);
    CHECK_INTEQUAL(pcm[38], mono_pcm[38]);
    CHECK_INTEQUAL(pcm[39], mono_pcm[39]);
    CHECK_FALSE(sound_decode_get_pcm(decoder, pcm, 1));

    sound_decode_close(decoder);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_loop_starts_past_end)
{
    SysFile *file;
    ASSERT(file = sys_file_open("testdata/sound/square.ogg"));
    SoundDecodeHandle *decoder;
    CHECK_TRUE(decoder = sound_decode_open_from_file(
                   SOUND_FORMAT_OGG, file, 0, sys_file_size(file),
                   1, 42, 10, 1));
    sys_file_close(file);
    CHECK_FALSE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4000);

    int16_t pcm[40];
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 40));
    CHECK_INTEQUAL(pcm[0], mono_pcm[0]);
    CHECK_INTEQUAL(pcm[1], mono_pcm[1]);
    CHECK_INTEQUAL(pcm[2], mono_pcm[2]);
    CHECK_INTEQUAL(pcm[3], mono_pcm[3]);
    CHECK_INTEQUAL(pcm[4], mono_pcm[4]);
    CHECK_INTEQUAL(pcm[5], mono_pcm[5]);
    CHECK_INTEQUAL(pcm[38], mono_pcm[38]);
    CHECK_INTEQUAL(pcm[39], mono_pcm[39]);
    CHECK_FALSE(sound_decode_get_pcm(decoder, pcm, 1));

    sound_decode_close(decoder);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_loop_zero_length)
{
    SysFile *file;
    ASSERT(file = sys_file_open("testdata/sound/square.ogg"));
    SoundDecodeHandle *decoder;
    CHECK_TRUE(decoder = sound_decode_open_from_file(
                   SOUND_FORMAT_OGG, file, 0, sys_file_size(file),
                   1, 2, 0, 1));
    sys_file_close(file);
    CHECK_FALSE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4000);

    int16_t pcm[40];
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 40));
    CHECK_INTEQUAL(pcm[0], mono_pcm[0]);
    CHECK_INTEQUAL(pcm[1], mono_pcm[1]);
    CHECK_INTEQUAL(pcm[2], mono_pcm[2]);
    CHECK_INTEQUAL(pcm[3], mono_pcm[3]);
    CHECK_INTEQUAL(pcm[4], mono_pcm[4]);
    CHECK_INTEQUAL(pcm[5], mono_pcm[5]);
    CHECK_INTEQUAL(pcm[38], mono_pcm[38]);
    CHECK_INTEQUAL(pcm[39], mono_pcm[39]);
    CHECK_FALSE(sound_decode_get_pcm(decoder, pcm, 1));

    sound_decode_close(decoder);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_enable_loop)
{
    SysFile *file;
    ASSERT(file = sys_file_open("testdata/sound/square.ogg"));
    SoundDecodeHandle *decoder;
    CHECK_TRUE(decoder = sound_decode_open_from_file(
                   SOUND_FORMAT_OGG, file, 0, sys_file_size(file),
                   0, 2, 3, 1));
    sys_file_close(file);
    CHECK_FALSE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4000);

    int16_t pcm[34];
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 3));
    CHECK_INTEQUAL(pcm[0], mono_pcm[0]);
    CHECK_INTEQUAL(pcm[1], mono_pcm[1]);
    CHECK_INTEQUAL(pcm[2], mono_pcm[2]);

    /* We should be able to enable the loop before we hit its endpoint. */
    sound_decode_enable_loop(decoder, 1);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 4));
    CHECK_INTEQUAL(pcm[0], mono_pcm[3]);
    CHECK_INTEQUAL(pcm[1], mono_pcm[4]);
    CHECK_INTEQUAL(pcm[2], mono_pcm[2]);
    CHECK_INTEQUAL(pcm[3], mono_pcm[3]);

    /* We should be able to disable the loop and play past its endpoint. */
    sound_decode_enable_loop(decoder, 0);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 2));
    CHECK_INTEQUAL(pcm[0], mono_pcm[4]);
    CHECK_INTEQUAL(pcm[1], mono_pcm[5]);

    /* Attempting to enable the loop when past its endpoint should not
     * cause an immediate loop. */
    sound_decode_enable_loop(decoder, 1);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 34));
    CHECK_INTEQUAL(pcm[32], mono_pcm[38]);
    CHECK_INTEQUAL(pcm[33], mono_pcm[39]);
    CHECK_FALSE(sound_decode_get_pcm(decoder, pcm, 1));

    /* If the decoder is subsequently reset, the loop flag should be
     * honored. */
    sound_decode_reset(decoder);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 9));
    CHECK_INTEQUAL(pcm[0], mono_pcm[0]);
    CHECK_INTEQUAL(pcm[1], mono_pcm[1]);
    CHECK_INTEQUAL(pcm[2], mono_pcm[2]);
    CHECK_INTEQUAL(pcm[3], mono_pcm[3]);
    CHECK_INTEQUAL(pcm[4], mono_pcm[4]);
    CHECK_INTEQUAL(pcm[5], mono_pcm[2]);
    CHECK_INTEQUAL(pcm[6], mono_pcm[3]);
    CHECK_INTEQUAL(pcm[7], mono_pcm[4]);
    CHECK_INTEQUAL(pcm[8], mono_pcm[2]);

    sound_decode_close(decoder);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_invalid_format)
{
    static const char * const files[] = {
        "testdata/sound/bad/empty-file.ogg",
        "testdata/sound/bad/not-vorbis.ogg",
        "testdata/sound/bad/4-channels.ogg",
    };
    for (int i = 0; i < lenof(files); i++) {
        SysFile *file;
        if (!(file = sys_file_open(files[i]))) {
            FAIL("sys_file_open(%s) failed: %s", files[i], sys_last_errstr());
        }
        if (sound_decode_open_from_file(
                SOUND_FORMAT_OGG, file, 0, sys_file_size(file), 0, 0, 0, 1)) {
            FAIL("sound_decode_open_from_file() for %s was not false as"
                 " expected", files[i]);
        }
        sys_file_close(file);
    }
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_corrupt_data)
{
    SysFile *file;
    ASSERT(file = sys_file_open("testdata/sound/bad/corrupt-data.ogg"));
    SoundDecodeHandle *decoder;
    CHECK_TRUE(decoder = sound_decode_open_from_file(
                   SOUND_FORMAT_OGG, file, 0, sys_file_size(file),
                   0, 0, 0, 1));
    sys_file_close(file);
    CHECK_FALSE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4000);

    int16_t pcm[1];
    CHECK_FALSE(sound_decode_get_pcm(decoder, pcm, 1));

    sound_decode_close(decoder);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_corrupt_data_looped)
{
    SysFile *file;
    ASSERT(file = sys_file_open("testdata/sound/bad/corrupt-data.ogg"));
    SoundDecodeHandle *decoder;
    CHECK_TRUE(decoder = sound_decode_open_from_file(
                   SOUND_FORMAT_OGG, file, 0, sys_file_size(file),
                   1, 0, 1, 1));
    sys_file_close(file);
    CHECK_FALSE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4000);

    int16_t pcm[1];
    /* Make sure this doesn't go into an infinite loop trying to loop over
     * a zero-length file. */
    CHECK_FALSE(sound_decode_get_pcm(decoder, pcm, 1));

    sound_decode_close(decoder);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_hole_in_data)
{
    SysFile *file;
    SoundDecodeHandle *decoder;
    int16_t pcm[1000];
    const int32_t hole_pos = 69380;

    /* We need >1 audio data packet to trigger the OV_HOLE logic, so we
     * use a 30-second square wave.  First make sure we have the correct
     * sample value for the original file. */
    ASSERT(file = sys_file_open("testdata/sound/square-long.ogg"));
    CHECK_TRUE(decoder = sound_decode_open_from_file(
                   SOUND_FORMAT_OGG, file, 0, sys_file_size(file),
                   0, 0, 0, 1));
    sys_file_close(file);
    CHECK_FALSE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4000);
    for (int32_t pos = 0; pos < hole_pos; pos += lenof(pcm)) {
        const int toread = ubound(hole_pos - pos, lenof(pcm));
        CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, toread));
    }
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 1));
#ifdef SIL_SOUND_USE_TREMOR
    CHECK_INTEQUAL(pcm[0], 9393);
#else
    CHECK_INTEQUAL(pcm[0], 9394);
#endif
    sound_decode_close(decoder);

    ASSERT(file = sys_file_open("testdata/sound/bad/holey-data.ogg"));
    CHECK_TRUE(decoder = sound_decode_open_from_file(
                   SOUND_FORMAT_OGG, file, 0, sys_file_size(file),
                   0, 0, 0, 1));
    sys_file_close(file);
    CHECK_FALSE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4000);
    for (int32_t pos = 0; pos < hole_pos; pos += lenof(pcm)) {
        const int toread = ubound(hole_pos - pos, lenof(pcm));
        CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, toread));
    }
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 1));
#ifdef SIL_SOUND_USE_TREMOR
    CHECK_INTEQUAL(pcm[0], 9392);
#else
    CHECK_INTEQUAL(pcm[0], 9393);
#endif
    sound_decode_close(decoder);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_decode_position_mismatch)
{
    SysFile *file;
    SoundDecodeHandle *decoder;
    int16_t pcm[1000];
    const int len = 120064;

    ASSERT(file = sys_file_open("testdata/sound/bad/granulepos-moves-back.ogg"));
    CHECK_TRUE(decoder = sound_decode_open_from_file(
                   SOUND_FORMAT_OGG, file, 0, sys_file_size(file),
                   0, 0, 0, 1));
    sys_file_close(file);
    CHECK_FALSE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4000);
    for (int32_t pos = 0; pos < len; pos += lenof(pcm)) {
        const int toread = ubound(len - pos, lenof(pcm));
        CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, toread));
    }
    CHECK_FALSE(sound_decode_get_pcm(decoder, pcm, 1));
    /* Make sure an appropriate warning is output as well. */
    CHECK_TRUE(strstr(test_DLOG_last_message, "corrupt"));

    sound_decode_close(decoder);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_decode_error)
{
    SysFile *file;
    ASSERT(file = sys_file_open("testdata/sound/square.ogg"));
    SoundDecodeHandle *decoder;
    CHECK_TRUE(decoder = sound_decode_open_from_file(
                   SOUND_FORMAT_OGG, file, 0, sys_file_size(file),
                   1, 2, 3, 1));
    sys_file_close(file);
    CHECK_FALSE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4000);

    int16_t pcm[40];
    /* Mess up the OggVorbis_File structure to force a read error. */
    OggVorbis_File *vf = (OggVorbis_File *)decoder->private;
    void *codec_setup_save = vf->vi->codec_setup;
    vf->vi->codec_setup = NULL;
    CHECK_FALSE(sound_decode_get_pcm(decoder, pcm, 1));
    vf->vi->codec_setup = codec_setup_save;
    /* Once the stream hits a hard error, it should not try to decode any
     * more data. */
    CHECK_FALSE(sound_decode_get_pcm(decoder, pcm, 1));

    /* This should reset the error flag. */
    sound_decode_reset(decoder);

    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 4));
    CHECK_INTEQUAL(pcm[0], mono_pcm[0]);
    CHECK_INTEQUAL(pcm[1], mono_pcm[1]);
    CHECK_INTEQUAL(pcm[2], mono_pcm[2]);
    CHECK_INTEQUAL(pcm[3], mono_pcm[3]);
    /* Fail on end-of-loop seek. */
    sound_decode_ogg_test_fail_next_seek();
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 1));
    CHECK_INTEQUAL(pcm[0], mono_pcm[4]);
    CHECK_FALSE(sound_decode_get_pcm(decoder, pcm, 1));

    sound_decode_reset(decoder);  // Fail on reset seek.
    CHECK_FALSE(sound_decode_get_pcm(decoder, pcm, 1));

    sound_decode_close(decoder);

    /* Fail on loop seek when the loop endpoint is past the end of the file. */
    ASSERT(file = sys_file_open("testdata/sound/square.ogg"));
    CHECK_TRUE(decoder = sound_decode_open_from_file(
                   SOUND_FORMAT_OGG, file, 0, sys_file_size(file),
                   1, 2, 40, 1));
    sys_file_close(file);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 39));
    sound_decode_ogg_test_fail_next_seek();
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 1));
    CHECK_FALSE(sound_decode_get_pcm(decoder, pcm, 1));
    sound_decode_close(decoder);

    return 1;
}

/*************************************************************************/
/*************************************************************************/

#endif  // SIL_SOUND_INCLUDE_OGG
#endif  // INCLUDE_TESTS

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
